/*
 * Copyright (c) 1996-2002 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

// $Id: pdwinnt.C,v 1.73 2003/01/03 21:57:37 bernat Exp $
#include <iomanip.h>
#include "dyninstAPI/src/symtab.h"
#include "common/h/headers.h"
#include "dyninstAPI/src/os.h"
#include "dyninstAPI/src/dyn_lwp.h"
#include "dyninstAPI/src/process.h"
#include "dyninstAPI/src/dyn_thread.h"
#include "dyninstAPI/src/stats.h"
#include "common/h/Types.h"
#include "dyninstAPI/src/showerror.h"
#include "dyninstAPI/src/instPoint.h"

#ifndef BPATCH_LIBRARY
#include "paradynd/src/main.h"
#include "paradynd/src/init.h"
#include "paradynd/src/perfStream.h" //SPLIT ccw 4 jun 2002
#endif

#ifdef BPATCH_LIBRARY
/* XXX This is only needed for emulating signals. */
#include "BPatch_thread.h"
#include "nt_signal_emul.h"
#endif

// prototypes of functions used in this file
static string GetLoadedDllImageName( process* p, const DEBUG_EVENT& ev );



//ccw  27 july 2000 : dummy methods to get the thing to compile before i add
//the remoteDevice : 29 mar 2001
#ifdef mips_unknown_ce2_11
#include <stdio.h> //for wprintf

void kludgeWCHAR(const char *str8, WCHAR *str16){
	int k;

	k=0;
	while(str8[k]!='\0'){
		str16[k] = str8[k];
		k++;
	}
	str16[k] = '\0';
	str16[k+1]='\0';

}

//int GetThreadContext(HANDLE hThread, w32CONTEXT *lpContext){
//	return 0;
//}

//int SetThreadContext(HANDLE hThread, const w32CONTEXT *lpContext){
//	return 0;
//}
#endif

extern bool isValidAddress(process *proc, Address where);
extern process *findProcess(int);

//HANDLE kludgeProcHandle;

void printSysError(unsigned errNo) {
    char buf[1000];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, errNo, 
		  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		  buf, 1000, NULL);
    fprintf(stderr, "*** System error [%d]: %s\n", errNo, buf);
    fflush(stderr);
}


// check if a file handle is for kernel32.dll
static bool kludge_isKernel32Dll(HANDLE fileHandle, string &kernel32Name) {
    static DWORD IndxHigh, IndxLow;
    static bool firstTime = true;
    BY_HANDLE_FILE_INFORMATION info;
    static string kernel32Name_;

    if (firstTime) {
	HANDLE kernel32H;
	firstTime = false;
	char sysRootDir[MAX_PATH+1];
	if (GetSystemDirectory(sysRootDir, MAX_PATH) == 0)
	  assert(0);
	kernel32Name_ = string(sysRootDir) + "\\kernel32.dll";
	kernel32H = CreateFile(kernel32Name_.c_str(), GENERIC_READ, 
			       FILE_SHARE_READ, NULL, OPEN_EXISTING, NULL, NULL);
	assert(kernel32H);
	if (!GetFileInformationByHandle(kernel32H, &info)) {
	    printSysError(GetLastError());
	    assert(0);
	}
	IndxHigh = info.nFileIndexHigh;
	IndxLow = info.nFileIndexLow;
	CloseHandle(kernel32H);
    }

    if (!GetFileInformationByHandle(fileHandle, &info))
	return false;

    if (info.nFileIndexHigh==IndxHigh && info.nFileIndexLow==IndxLow) {
      kernel32Name = kernel32Name_;
      return true;
    }
    return false;
}


void dumpMem(process *p, void * addr, unsigned nbytes) {
    unsigned char *buf = new unsigned char[nbytes];
    memset(buf, 0, nbytes);
    function_base *f;
    assert(buf);

    if (f = p->findFuncByAddr((Address)addr))
    {
        printf("Function %s, addr=0x%lx, sz=%d\n", 
                f->prettyName().c_str(),
                f->getAddress(p),
                f->size());
    }
    p->readDataSpace((void *)((unsigned)addr-32), nbytes, buf, true);
    printf("## 0x%lx:\n", (unsigned)addr-32);
    for (unsigned u = 0; u < nbytes; u++)
    {
	    printf(" %x", buf[u]);
    }
    printf( "\n" );
    p->readDataSpace(addr, nbytes, buf, true);
    printf("## 0x%lx:\n", addr);
    for (unsigned u1 = 0; u1 < nbytes; u1++)
    {
	    printf(" %x", buf[u1]);
    }
    printf( "\n" );
}



//
// walkStackFrame
//
// Try to walk one more frame in the stack of the indicated process.
// Assumes that the STACKFRAME argument is either the result
// from a previous StackWalk call, or has been initialized 
// to the context from a GetThreadContext call.
//
inline
BOOL
walkStackFrame( HANDLE hProc, HANDLE hThread, STACKFRAME* psf )
{
	return StackWalk( IMAGE_FILE_MACHINE_I386,
                        hProc,
                        hThread,
		                psf,
                        NULL,
		                NULL,
                        SymFunctionTableAccess,
                        SymGetModuleBase,
		                NULL);
}


//
// findFunctionFromAddress
// 
// Finds the function corresponding to the given address.
// Handles situations where the address is within instrumentation,
// and situations where the function has been relocated.
//
function_base*
findFunctionFromAddress( process* proc, Address addr )
{
    function_base* fp = NULL;

    instPoint* ip = proc->findInstPointFromAddress(addr);
    if( ip != NULL )
    {
        // the address is within a tramp
        // find the function in which the tramp was installed
        fp = (function_base*)( ip->iPgetFunction() );
    }
    else
    {
        // the address isn't within a tramp
        // check if it is a known function (relocated or otherwise)
        fp = (function_base*)( proc->findFuncByAddr( addr ) );
    }
    return fp;
}



#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
//
// process::walkStackFromFrame
//
// 8OCT02: this should now pay proper attention to threads,
//         though a better method than handing in a thread to
//         kickstart the stackwalk is probably a good idea -- Bernat
//
// Note that we have *not* been able to find a mechanism that 
// can perform stack walks reliably for apps compiled with the FPO 
// (frame pointer omission) optimization.  Even if the FPO data is
// available, the Win32 StackWalk call either skips functions or bottoms
// out the stack too early.  Unfortunately, this means that we require
// that the app under study *not* use the FPO optimization.  (Specify /Oy- 
// for the VC++ compiler to turn this optimization off.)
//

bool process::walkStackFromFrame(Frame currentFrame, pdvector<Frame> &stackWalk)
{
#ifndef BPATCH_LIBRARY
    startTimingStackwalk();
#endif
    
#ifdef DEBUG_STACKWALK
    cout << "\n<stack>" << endl;
#endif // DEBUG_STACKWALK
    
    if (status_ == running) {
        cerr << "Error: stackwalk attempted on running process" << endl;
#ifndef BPATCH_LIBRARY
      stopTimingStackwalk();
#endif
      return false;
    }
    
    // establish the current execution context
    CONTEXT cont;
    HANDLE hThread = (HANDLE)(currentFrame.getLWP()->get_fd());
    
    cont.ContextFlags = CONTEXT_FULL;
    if (!GetThreadContext(hThread, &cont))
    {
        cerr << "walkStack: GetThreadContext failed:" << GetLastError()
             << endl;

#ifndef BPATCH_LIBRARY
	stopTimingStackwalk();
#endif
        return false;
    }

    STACKFRAME sf;
    ZeroMemory( &sf, sizeof(STACKFRAME) );
    sf.AddrPC.Offset = cont.Eip;
    sf.AddrPC.Mode = AddrModeFlat;
    sf.AddrFrame.Offset = cont.Ebp;
    sf.AddrFrame.Mode = AddrModeFlat;
    sf.AddrStack.Offset = cont.Esp;
    sf.AddrStack.Mode = AddrModeFlat;
    sf.AddrReturn.Mode = AddrModeFlat;

    // walk the stack, frame by frame
    // we use the Win32 StackWalk function to automatically
    // handle compiler optimizations, especially FPO optimizations
    bool done = false;

    while( !done ) {
        STACKFRAME saved_sf = sf;
	BOOL walked;
     	ADDRESS patchedAddrReturn;
       	ADDRESS patchedAddrPC;
       	instPoint* ip = NULL;
       	function_base* fp = NULL;

       	// set defaults for return address and PC
       	patchedAddrReturn = saved_sf.AddrReturn;
       	patchedAddrPC = saved_sf.AddrPC;

	// try to step through the stack using the current information
      	walked = walkStackFrame((HANDLE)procHandle_,  hThread, &sf);

	if( !walked && (GetLastError() == ERROR_INVALID_ADDRESS) ) {

          // try to patch the return address, in case it is outside
          // of the original text of the process.  It might be outside
          // the original text if the function has been relocated, or
  	  // if it represents a return from a call instruction that
          // has been relocated to a base tramp.
          // we first try to patch the return address only, because it
          // is most likely that the return address only is out of the
          // original text space.  Once we process the first stack
          // frame, it appears that the StackWalk function transfers 
          // the return address from the STACKFRAME struct to the PC,
          // so once we've gotten a valid return address in a 
          // STACKFRAME, the PC will be within the original text after
          // the StackWalk
          sf = saved_sf;
          fp = findFunctionFromAddress( this, sf.AddrReturn.Offset );

	  if( fp != NULL ) {
      	    // because StackWalk seems to support it, we simply use 
       	    // the address of the function itself rather than
       	    // trying to do the much more difficult task of finding
       	    // the original address of the relocated instruction
       	    patchedAddrReturn.Offset = fp->addr();
       	    sf.AddrReturn = patchedAddrReturn;
	  }

          // retry the stack step
          walked = 
              walkStackFrame((HANDLE)procHandle_, hThread, &sf);
	}

	if( !walked && (GetLastError() == ERROR_INVALID_ADDRESS) ) {

       	  // patching the return address alone didn't work.
       	  // try patching the return address and the PC
       	  sf = saved_sf;
          fp = findFunctionFromAddress( this, sf.AddrPC.Offset );

       	  if( fp != NULL ) {

       	    // because StackWalk seems to support it, we simply use 
       	    // the address of the function itself rather than
       	    // trying to do the much more difficult task of finding
       	    // the original address of the relocated instruction
       	    patchedAddrPC.Offset = fp->addr();
       	    sf.AddrPC = patchedAddrPC;
       	  }

       	  // use the patched return address we calculated above
       	  sf.AddrReturn = patchedAddrReturn;

       	  // retry the stack step
       	  walked = 
              walkStackFrame((HANDLE)procHandle_, hThread, &sf);
       	}

       	if( !walked && (GetLastError() == ERROR_INVALID_ADDRESS) )
       	{

       	  // patching both addresses didn't work
       	  //
       	  // try patching the PC only
      	  sf = saved_sf;
       	  sf.AddrPC = patchedAddrPC;

      	  // retry the stack step
       	  walked = 
              walkStackFrame((HANDLE)procHandle_, hThread, &sf);
       	}


	// by now we've tried all of our tricks to handle the stack 
        // frame if we haven't succeeded by now, we're not going to be
        // able to handle this frame
      	if( walked ) {
	  
	  Address pc = NULL;
	  
	  // save the PC for this stack frame
	  // make sure we use the original address in case
	  // it was outside the original text of the process
	  if( saved_sf.AddrReturn.Offset == 0 ) {
	    
	    // this was the first stack frame, so we had better
	    // use the original PC
	    pc = saved_sf.AddrPC.Offset;
	    
	  } else {
	    
	    // this was not the first stack frame
	    // use the original return address
	    pc = saved_sf.AddrReturn.Offset;
	  }
	
	  stackWalk.push_back(Frame(pc, 0, getPid(), 
				    currentFrame.getThread(), 
				    currentFrame.getLWP(), 
				    false));

#ifdef DEBUG_STACKWALK
	  cout << "0x" << setw(8) << setfill('0') << hex << pc << ": ";
	  
	  if( fp != NULL ) {
	    cout << fp->prettyName();
	  } else {
	    cout << "<unknown>";
	  }
	  
	  if( sf.AddrPC.Offset != pc ) {
	    cout << " (originally 0x" << setw(8) << setfill('0') 
		 << hex << sf.AddrPC.Offset << ")";
	  }
	  cout << endl;
#endif
	  
	} else {
	  // we tried everything we know to recover - we'll have 
	  // to fail
	  done = true;
	}
	
    }

    // Terminate stack walk with an empty frame
    stackWalk.push_back(Frame());
#ifndef BPATCH_LIBRARY
    stopTimingStackwalk();
#endif

    return true;
}
#endif

#if defined(mips_unknown_ce2_11) //ccw 6 feb 2001 : 29 mar 2001
//ccw 6 feb 2001 : windows CE does not have the NT walkStack function
//so we use this one.

bool process::walkStackFromFrame(Frame currentFrame, pdvector<Frame> &stackWalk)
{
    if (status_ == running) {
        cerr << "Error: stackwalk attempeded on running process" << endl;
        return false;
    }  
    Address spOld = 0xffffffff;
    
    while (!currentFrame.isLastFrame()) {
        Address spNew = currentFrame.getSP(); // ccw 6 feb 2001 : should get SP?
        
        // successive frame pointers might be the same (e.g. leaf functions)
        if (spOld < spNew) {
            return false;            
        }
        
        spOld = spNew;
        
        Address next_pc = currentFrame.getPC();
        stackWalk.push_back(currentFrame);
        
        //ccw 6 feb 2001 : at this point, i need to use the 
        //list of functions parsed from the debug symbols to
        //determine the frame size for each function and find the 
        //return value for each (which is the previous fir value)
        currentFrame = currentFrame.getCallerFrame(this); 
        
    }    
    stackWalk.push_back(currentFrame);
    return true;    
}
#endif


//SPLIT ccw 4 jun 2002
#if !defined(BPATCH_LIBRARY)
//just a dummy so it compiles.
bool process::dlopenPARADYNlib(){ 
	return true;
}
#endif

/* 
   Loading libDyninstRT.dll

   We load libDyninstRT.dll dynamically, by inserting code into the
   application to call LoadLibraryA. We don't use the inferior RPC
   mechanism from class process because it already assumes that
   libdyninst is loaded (it uses the inferior heap).
   Instead, we use a simple inferior call mechanism defined below
   to insert the code to call LoadLibraryA("libdyninstRT.dll").
 */

Address loadDyninstDll(process *p, char Buffer[LOAD_DYNINST_BUF_SIZE]) {
    Address codeBase = p->getImage()->codeOffset();
	Address hackLoadLibAddr = 0x01f9ac30; //ccw 2 feb 2001 : HACK

    Address LoadLibBase;
    Address LoadLibAddr;
    Symbol sym;
 	//ccw 22 aug 2000 : this function hand codes x86 op codes!!!!
	//change these to mips instructions!!!
#ifdef mips_unknown_ce2_11 //ccw 14 aug 2000 : 29 mar 2001

	//ccw 2 feb 2001
	//this is a terrible hack.  i know that LoadLibrary is 0xac30 bytes from
	//the base address of coredll.dll from emprical evidence.  i have no
	//debug symbols for coredll.dll to prove this dynamically though....
	
	for(int i=0; i< (p->sharedObjects())->size();i++){
		
		if((*p->sharedObjects())[i]->getName() == "coredll.dll" ){
			hackLoadLibAddr = (*p->sharedObjects())[i]->getBaseAddress() + 0x0000bf1c;
			//0x0000ac30; //HACK 24 apr 2001
		}
	}

#if 0 
	//ccw 2 feb 2001 : this is taken out because the above
	//loop determines the correct address for LoadLibrary
	//this statement requires LoadLibrary to show up in
	//the map file of the mutatee, which only happens if
	//the mutatee explicitly calls LoadLibrary
	if (!p->getSymbolInfo("LoadLibraryW", sym, LoadLibBase)) {

	 if( !p->getSymbolInfo( "__imp_LoadLibraryW", sym, LoadLibBase )) //ccw 1 oct 2000, not a good idea
         {
	        printf("unable to find function LoadLibrary\n");
	        assert(0);
        }
    }
#endif 

#else
	if (!p->getSymbolInfo("_LoadLibraryA@4", sym, LoadLibBase)) {
        if( !p->getSymbolInfo( "_LoadLibraryA", sym, LoadLibBase ))
        {
	        printf("unable to find function LoadLibrary\n");
	        assert(0);
        }
    }
#endif

#ifndef mips_unknown_ce2_11 //ccw 2 feb 2001
    LoadLibAddr = sym.addr() + LoadLibBase;
    assert(LoadLibAddr);
#endif

    char ibuf[LOAD_DYNINST_BUF_SIZE];
    char *iptr = ibuf;
	memset(ibuf, '\0', LOAD_DYNINST_BUF_SIZE);//ccw 25 aug 2000

#ifdef mips_unknown_ce2_11 //ccw 22 aug 2000
	//the following are the instructions to load libdyninstRT_API.dll
	
	//ccw 2 oct 2000
	//the instructions we need to generate are:
	// lui $a0,hi16(codeBase)  # codeBase contains libBuf
	// ori $a0,lo16(codeBase)
	// lui $t0, hi16(hackLoadLibAddr)
	// ori $t0, lo16(hackLoadLibAddr)
	// jalr $t0
	// nop
	// nop
	// break
	// nop
	// nop
	// nop
	// where libBuf is the location of the string containing the name of the dll
	// in the remote process's memory

	WCHAR libBuf[] = L"\\libdyninstAPI_RT.dll";
	BYTE instBuff[4]; //buffer for the instruction to write to memory;

	//lay the name of the DLL to load into memory, UNICODE!
	assert(memcpy(iptr, (char*)libBuf, 42));
	iptr +=42;
	memset(iptr, '\0', 2);
	iptr +=2;

	//LUI //ccw 2 oct 2000
	instBuff[0]=((char*) &codeBase)[2];
	instBuff[1]=((char*) &codeBase)[3];
	instBuff[2]=0x04;
	instBuff[3]=0x3C;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//ORI //ccw 2 oct 2000
	instBuff[0]=((char*) &codeBase)[0];
	instBuff[1]=((char*) &codeBase)[1];
	instBuff[2]=0x84;
	instBuff[3]=0x34;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//LUI //ccw 2 feb 2001
	instBuff[0]=((char*) &hackLoadLibAddr)[2];
	instBuff[1]=((char*) &hackLoadLibAddr)[3];
	instBuff[2]=0x08;
	instBuff[3]=0x3C;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//ORI //ccw 2 feb 2001
	instBuff[0]=((char*) &hackLoadLibAddr)[0];
	instBuff[1]=((char*) &hackLoadLibAddr)[1];
	instBuff[2]=0x08;
	instBuff[3]=0x35;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//JALR
	instBuff[0]=0x09;
	instBuff[1]=0xf8;
	instBuff[2]=0x00;
	instBuff[3]=0x01;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;


	//NOP
	instBuff[0]=0x00;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//NOP
	instBuff[0]=0x00;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;


	//DebugBreak();
	instBuff[0]=0x0D;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;


	//ccw 25 aug 2000 : throw some NOPs after the debugBreak just in case!
	//NOP
	instBuff[0]=0x00;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;
	
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

#else

    // push nameAddr ; 5 bytes
    *iptr++ = (char)0x68; 
    *(int *)iptr = codeBase + 14; 
    iptr += sizeof(int);

    // call LoadLibrary ; 5 bytes
    *iptr++ = (char)0xe8;
    // offset relative to next instruction (address codeBase+10)
    *(int *)iptr = LoadLibAddr - (codeBase + 10);
    iptr += sizeof(int);

    // add sp, 4
    *iptr++ = (char)0x83; *iptr++ = (char)0xc4; *iptr++ = (char)0x04;

    // int3
    *iptr++ = (char)0xcc;

    /*if (!process::dyninstRT_name.length()) ccw 28 aug 2002 ALWAYS CHECK ENV VAR*/
        // check for an environment variable
	// SPLIT ccw 4 jun 2002
#if defined( BPATCH_LIBRARY ) || 1  // dyninstAPI loads a different run-time library
        process::dyninstRT_name = getenv("DYNINSTAPI_RT_LIB");
#endif

    if (!process::dyninstRT_name.length())
        // if environment variable unset, use the default name/strategy
#if defined( BPATCH_LIBRARY ) || 1 //SPLIT ccw 4 jun 2002
        process::dyninstRT_name = "libdyninstAPI_RT.dll";
#endif
	
    // make sure that directory separators are what LoadLibrary expects
    strcpy(iptr, process::dyninstRT_name.c_str());
    for (unsigned int i=0; i<strlen(iptr); i++)
        if (iptr[i]=='/') iptr[i]='\\';
#endif

    p->readDataSpace((void *)codeBase, LOAD_DYNINST_BUF_SIZE, Buffer, false);
    p->writeDataSpace((void *)codeBase, LOAD_DYNINST_BUF_SIZE, ibuf);

#ifdef mips_unknown_ce2_11 //ccw 22 aug 2000
	//add to process, then to remoteDevice
   return codeBase+44;//this is for the offset of the string!
#else
	return codeBase;
#endif
}

//SPLIT ccw 4 jun 2002 
/* 
   Loading libDyninstRT.dll

   We load libDyninstRT.dll dynamically, by inserting code into the
   application to call LoadLibraryA. We don't use the inferior RPC
   mechanism from class process because it already assumes that
   libdyninst is loaded (it uses the inferior heap).
   Instead, we use a simple inferior call mechanism defined below
   to insert the code to call LoadLibraryA("libdyninstRT.dll").
 */
#ifndef BPATCH_LIBRARY
#if !defined(mips_unknown_ce2_11)
Address loadParadynDll(process *p ) {
    Address codeBase = p->getImage()->codeOffset();
	Address hackLoadLibAddr = 0x01f9ac30; //ccw 2 feb 2001 : HACK

    Address LoadLibBase;
    Address LoadLibAddr;
    Symbol sym;

	if (!p->getSymbolInfo("_LoadLibraryA@4", sym, LoadLibBase)) {
        if( !p->getSymbolInfo( "_LoadLibraryA", sym, LoadLibBase ))
        {
	        printf("unable to find function LoadLibrary\n");
	        assert(0);
        }
    }

    LoadLibAddr = sym.addr() + LoadLibBase;
    assert(LoadLibAddr);

    char ibuf[LOAD_DYNINST_BUF_SIZE];
    char *iptr = ibuf;
	memset(ibuf, '\0', LOAD_DYNINST_BUF_SIZE);//ccw 25 aug 2000

    // push nameAddr ; 5 bytes
    *iptr++ = (char)0x68; 
    *(int *)iptr = codeBase + 14; 
    iptr += sizeof(int);

    // call LoadLibrary ; 5 bytes
    *iptr++ = (char)0xe8;
    // offset relative to next instruction (address codeBase+10)
    *(int *)iptr = LoadLibAddr - (codeBase + 10);
    iptr += sizeof(int);

    // add sp, 4
    *iptr++ = (char)0x83; *iptr++ = (char)0xc4; *iptr++ = (char)0x04;

    // int3
    *iptr++ = (char)0xcc;

    //    if (!process::paradynRT_name.length())
    // check for an environment variable
    process::paradynRT_name = getenv("PARADYN_LIB");

    if (!process::paradynRT_name.length())
        // if environment variable unset, use the default name/strategy
        process::paradynRT_name = "libdyninstRT.dll";
	
    // make sure that directory separators are what LoadLibrary expects
    strcpy(iptr, process::paradynRT_name.c_str());
    for (unsigned int i=0; i<strlen(iptr); i++)
        if (iptr[i]=='/'){
		 iptr[i]='\\';
	}
    	p->writeDataSpace((void *)codeBase, LOAD_DYNINST_BUF_SIZE, ibuf);

	return codeBase;
}
#endif
#endif

// osTraceMe is not needed in Windows NT
void OS::osTraceMe(void) {}


#ifndef BPATCH_LIBRARY
void checkProcStatus() {
    int wait_status;
    extern void doDeferredRPCs();

    process::waitProcs(&wait_status);
    doDeferredRPCs();
}
#endif


// ccw 2 may 2001 win2k fixes
// when you launch a process to be debugged with win2k (as in createProcess)
// the system sends you back at least two debug events before starting the
// process.  a debug event is also sent back for every dll that is loaded
// prior to starting main(), these include ntdll.dll and kernel32.dll and any
// other dlls the process needs that are not loaded with an explicit call
// to LoadLibrary().
//
// dyninst catches the first debug event (CREATE_PROCESS) and initializes
// various process specific data structures.  dyninst catches the second
// debug event (an EXCEPTION_DEBUG_EVENT) and used this event as a trigger to
// put in the bit of code that forced the mutatee to load
// libdyninstAPI_RT.dll.  In win2k, this does not work.  The bit of code is
// run and the trailing DebugBreak is caught and handled but the Dll will not
// be loaded.  The EXCEPTION_DEBUG_EVENT must be handled and continued from
// before LoadLibrary will perform correctly.
//
// the fix for this is to handle this EXCEPTION_DEBUG_EVENT, and put a
// DebugBreak (0xcc) at the beginning of main() in the mutatee.  catching
// that DebugBreak allows dyninst to write in the bit of code used to load
// the libdyninstAPI_RT.dll.
//
// after this, dyninst previously instrumented the mutatee to force the
// execution of DYNINSTinit() in the dll.  in order to take out this bit of
// complexity, the DllMain() function in the dll, which is run upon loading
// the dll, is used to automatically call DYNINSTinit().
//
// DYNINSTinit() takes two parameters, a flag denoting how dyninst attached
// to this process and the pid of the mutator.  These are passed from the
// mutator to the mutatee by finding a variable in the dll and writing the
// correct values into the mutatee's address space.  When a Dll is loaded, a
// LOAD_DLL debug event is thrown before the execution of DllMain(), so
// dyninst catches this event, writes the necessary values into the mutatee
// memory, then lets DllMain() call DYNINSTinit().  the DebugBreak() at the
// end of DYNINSTinit() is now removed for NT/win2K
//
// the bit of code inserted to load the dll fires a DebugBreak() to signal
// that it is done. dyninst catches this, patches up the code that was used
// to load the dll, replaces what was overwritten in main() and resets the
// instruction pointer (EIP) to the beginning of main().

//ccw 28 aug 2002 : the following five vars have moved to Process so
// we can properly load more than one mutatee
 
//ccw 30 apr 2001 : used to signal-we have seen the first EXCEPTION_DEBUG_EVENT
//int secondBkpt = 0; 

//ccw 2 may 2001 : used to track where we wrote the LoadLibrary code
//int mungeAddr = 0; 
//byte savedOpCode[256]; //ccw 2 may 2001 : the op code we overwrite in main()
//byte newOpCode[256];   // ccw 27 june 2001 : the op code we add in main()
//ccw 6 july 2001 the two above arrays are 256 bytes because writing one byte
//does not always cause the instruction cache to be reloaded.

//Address p->mainAddr=0 ; //ccw 2 may 2001 : the address of main()

//ccw 2 may 2001 : the following function sets the varible name to the value
//given.  returns true if it is successful.
bool setVariable(process* p, string name, int value){
    	string full_name = string("_") + string(name);
	Symbol syminfo;
	string short_name(name);
	if (!p->getSymbolInfo(short_name, syminfo)) {
		if (!p->getSymbolInfo(full_name, syminfo)) {
		    assert(0);
		}
	}else{
		full_name = short_name;
	}

	if( syminfo.type() == Symbol::PDST_FUNCTION)
   		assert(0); 
	Address tmpAddr = syminfo.addr();
	int oldvalue;	
  	/*
        p->readDataSpace((void*)tmpAddr, sizeof(int), (void*) &oldvalue,true);
	printf("\n\n READ ORIGINAL VAR %s %x %d\n", full_name.c_str(), tmpAddr, oldvalue);
	*/
	int ret= p->writeDataSpace((void*)tmpAddr, sizeof(int), (void*) &value);
	/*
	printf(" SET VAR %s %x %d %d\n", full_name.c_str(), tmpAddr, value, ret);
	p->readDataSpace((void*)tmpAddr, sizeof(int), (void*) &value,true);
	printf(" READ VAR %s %x %d %d\n", full_name.c_str(), tmpAddr, value, ret);
	fflush(stdout);
	*/
	return ret;
}

#ifdef mips_unknown_ce2_11
int secondDLL = 0; //ccw 24 oct 2000 : 28 mar 2001
#endif
DWORD continueType = DBG_CONTINUE; //ccw 25 oct 2000 : 28 mar 2001


/*
   wait for inferior processes to terminate or stop.
*/
#ifdef BPATCH_LIBRARY
int process::waitProcs(int *status, bool block) {
#else
int process::waitProcs(int *status) {
#endif
    DEBUG_EVENT debugEv;
    process *p;
#ifdef BPATCH_LIBRARY
    *status = 0;
#endif

    // We wait for 1 millisecond here. On the Unix platforms, the wait
    // happens on the select in controllerMainLoop. But on NT, because
    // we have to handle traps, we set the timeout on the select as 0,
    // so that we can check for traps quickly
#ifdef BPATCH_LIBRARY
    DWORD milliseconds;
    if (block) milliseconds = INFINITE;
    else milliseconds = 1;

#if defined(mips_unknown_ce2_11) //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteWaitForDebugEvent(&debugEv, milliseconds))
#else
    if (!WaitForDebugEvent(&debugEv, milliseconds))
#endif    

#else

    if (!WaitForDebugEvent(&debugEv, 1))

#endif
	return 0;

    p = findProcess(debugEv.dwProcessId);
    if (p == NULL) {
	/* 
	   this case can happen when we create a process, but then are
	   unable to parse the symbol table, and so don't complete the
	   creation of the process. We just ignore the event here.  
	*/

#if defined(mips_unknown_ce2_11) //ccw 28 july 2000 : 29 mar 2001
	BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			   DBG_CONTINUE);
#else
		ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			   DBG_CONTINUE);

#endif
	return 0; 
    }
#if defined(mips_unknown_ce2_11 )
	continueType = DBG_CONTINUE;
#endif
    switch (debugEv.dwDebugEventCode) {
    case EXCEPTION_DEBUG_EVENT: {
	DWORD exnCode = debugEv.u.Exception.ExceptionRecord.ExceptionCode;
	switch(exnCode) {
	case EXCEPTION_BREAKPOINT: {
	    /*printf("Debug breakpoint exception, %d, addr = %x\n", 
	           debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
	           debugEv.u.Exception.ExceptionRecord.ExceptionAddress);*/ //PRINTF
	    
#if defined(mips_unknown_ce2_11) //ccw 20 mar 2001 : 29 mar 2001
		continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 25 oct 2000
#endif

#if !defined(mips_unknown_ce2_11)
	    	if(!p->secondBkpt && !p->wasCreatedViaAttach()){
			// if this was created via attach then
			// we are not at main() so no reason to
			// put a breakpoint there, we are already
			// at a state that will allow loadLibrary() to
			// succeed. 
			//ccw this is the breakpoint BEFORE main();
			//i need to insert the 0xcc into main() now.
			//byte debugOpCode = 0xcc;
	 	    	p->savedRegs = p->getDefaultLWP()->getRegisters();
			function_base *mainFunc;
			//DebugBreak();//ccw 14 may 2001  
			if (!((mainFunc = p->findOneFunction("main")))){
				if(!(mainFunc = p->findOneFunction("_main"))){
					if(!(mainFunc = p->findOneFunction("WinMain"))){
						if(!(mainFunc = p->findOneFunction("_WinMain"))){
							assert(0);
						}
					}
				}

			}
			p->mainAddr = mainFunc->addr();
			//ccw save what we overwrite for later.
			p->readDataSpace_((void*) (p->mainAddr), 256, (void*)p->savedOpCode);

			// insert the breakpoint into the code we just read
			CopyMemory( p->newOpCode, p->savedOpCode, 256 );
			p->newOpCode[0] = (char) 0xcc;

			// write the modified code sequence back
			p->writeDataSpace((void*) (p->mainAddr), 256, (void*)p->newOpCode);
			p->flushInstructionCache_( (void*)p->mainAddr, 256 );

			p->secondBkpt=1;
			break;
		}
#endif


	    if (!p->reachedFirstBreak
#if !defined(BPATCH_LIBRARY) //ccw 4 jun 2002 SPLIT
			    || !p->hasLoadedParadynLib
#endif
			    ) {
		int inst_offset=  0xd; 
		if (!p->hasLoadedDyninstLib && !p->isLoadingDyninstLib) {
			//ccw 11 july 2001: IS THIS TRUE?
			//ccw 28 feb 2001
			//for MIPS/WINCE this never happens here, it is done in the
			//LOAD_DLL exception handler.  this is because of the
			//bug in which WINCE only passes one debug break when
			//it starts a process (unlike NT which gives two).
			//more explaination below in LOAD_DLL
			//DebugBreak();//ccw 30 jan 2001
			continueType = DBG_CONTINUE;//ccw 25 oct 2000 : 29 mar 2001

			//printf(" Loading dyninst lib\n");//PRINTF
		    Address addr = loadDyninstDll(p, p->savedData);
	   	    p->mungeAddr = addr; //ccw 15 june 2001
		    p->savedRegs = p->getDefaultLWP()->getRegisters();
		    p->getDefaultLWP()->changePC(addr, NULL);
		    //p->LoadDyninstTrapAddr = addr + 0xd;
		    p->isLoadingDyninstLib = true;
		    break;
		} else if (p->isLoadingDyninstLib
#if defined(mips_unknown_ce2_11)
			){
			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 25 oct 2000
#else
			&& ((int) p->mungeAddr + inst_offset) ==
			(int) debugEv.u.Exception.ExceptionRecord.ExceptionAddress) {
#endif
			//NOTE the above test is different for CE/NT
			//
			//ccw 2 may 2001 the above test is changed so that we only
			// respond to events after
			// the load library code has completed.  (we know that the 
			// DebugBreak 0xd bytes from 
			// the beginning of where we inserted code signals that we
			// have finished load library) 

		    p->isLoadingDyninstLib = false;
		    p->hasLoadedDyninstLib = true;
 			//printf(" caught breakpoint after load of DYNINST lib\n");//PRINTF
			
#if defined(BPATCH_LIBRARY)  //ccw 4 jun 2002 SPLIT
			
#if !defined(mips_unknown_ce2_11)
	 		//ccw 25 june 2001 NEED TO FLUSH ICACHE here
			p->flushInstructionCache_((void*) ((w32CONTEXT*) p->savedRegs)->Eip,0x100);
#endif

			if(p->mainAddr){
			    // patch main() back to its original form
			    p->writeDataSpace((void*) (p->mainAddr),16,(void*) p->savedOpCode); //ccw 2 may 2001
			    // reset the IP to run what we just inserted.
			    ((w32CONTEXT*) p->savedRegs)->Eip-=1; //ccw 2 may 2001
			}

		    p->getDefaultLWP()->restoreRegisters(p->savedRegs);
		    delete p->savedRegs;
		    p->writeDataSpace((void *)p->getImage()->codeOffset(),
				      LOAD_DYNINST_BUF_SIZE,
				      (void *)p->savedData);
#if !defined(mips_unknown_ce2_11)		    
			if(p->mainAddr){
				//ccw 25 june 2001 NEED TO FLUSH ICACHE here
				//p->flushInstructionCache_((void*) ((CONTEXT*) p->savedRegs)->Eip,0x100);
				p->readDataSpace_((void*)(p->mainAddr), 256, (void*) p->newOpCode);
				p->writeDataSpace((void*) (p->mainAddr),256,(void*) p->savedOpCode); //ccw 2 may 2001
				//technically, we only need to writeDataSpace once, up at the previous
				// if statement. BUT writing once does not seem to always flush the 
				// icache, and neither does FlushInsturctionCache()
				// the readDataSpace_ is completely spurrious but is here for the same
				// reason, to get the icache updated correctly with the old op code before
				// we continue. 
			}
#endif
#endif


#if !defined(BPATCH_LIBRARY) //SPLIT ccw 4 jun 2002

		}else if(p->isLoadingParadynLib
			&& ((int) p->mungeAddr + inst_offset) ==
			(int) debugEv.u.Exception.ExceptionRecord.ExceptionAddress) {
			
		    	p->hasLoadedParadynLib = true;
		    	p->isLoadingParadynLib = false;
			//printf(" caught breakpoint after load of paradyn lib\n");//PRINTF
	 		//ccw 25 june 2001 NEED TO FLUSH ICACHE here
			p->flushInstructionCache_((void*) ((w32CONTEXT*) p->savedRegs)->Eip,0x100);
			if(p->mainAddr){
			    // patch main() back to its original form
			    p->writeDataSpace((void*) (p->mainAddr),16,(void*) p->savedOpCode); //ccw 2 may 2001
			    // reset the IP to run what we just inserted.
			    ((w32CONTEXT*) p->savedRegs)->Eip-=1; //ccw 2 may 2001
			}

			p->getDefaultLWP()->restoreRegisters(p->savedRegs);
			delete p->savedRegs;
			p->writeDataSpace((void *)p->getImage()->codeOffset(),
				      LOAD_DYNINST_BUF_SIZE,
				      (void *)p->savedData);
			if(p->mainAddr){
				//ccw 25 june 2001 NEED TO FLUSH ICACHE here
				//p->flushInstructionCache_((void*) ((CONTEXT*) p->savedRegs)->Eip,0x100);
				p->readDataSpace_((void*)(p->mainAddr), 256, (void*) p->newOpCode);
				p->writeDataSpace((void*) (p->mainAddr),256,(void*) p->savedOpCode); //ccw 2 may 2001
				//technically, we only need to writeDataSpace once, up at the previous
				// if statement. BUT writing once does not seem to always flush the 
				// icache, and neither does FlushInsturctionCache()
				// the readDataSpace_ is completely spurrious but is here for the same
				// reason, to get the icache updated correctly with the old op code before
				// we continue. 
			}

			p->status_ = stopped;
			p->pause_();
			
			/*if(p->createdViaAttach){ //ccw 7 jun 2002
				p->handleCompletionOfpDYNINSTinit(true);
			}*/
			break;

#endif			
		}else{
#if !defined(BPATCH_LIBRARY) // ccw 5 jun 2002 : SPLIT
			if(p->isLoadingParadynLib ){ //ccw 5 jun 2002 SPLIT
			    p->status_ = stopped;
			    p->pause_();
			    int result = p->procStopFrompDYNINSTinit();
			    //printf(" RESULT from pDYNINST init %d \n", result);//PRINTF
			    assert(result >= 0 && result <= 2);		
			    //fflush(stdout);//PRINTF
			    p->status_ = running;
			    p->continueProc_();
			}
#endif
			break;  //ccw 20 june 2001 this is the DYNINSTinit breakpoint
		}
		p->reachedFirstBreak = true;
      
		if (!p->createdViaAttach) {
		    int pid = p->getPid();
		    p->status_ = stopped;

		    if (!p->initDyninstLib()) {
			OS::osKill(pid);
			handleProcessExit(p, -1);
#ifdef BPATCH_LIBRARY
			*status = SIGEM_SIGNALED | SIGTERM;
			return p->getPid();
#else
			return 0;
#endif
		    }

		    string buffer = string("PID=") + string(pid);
		    buffer += string(", passed trap at start of program");
		    statusLine(buffer.c_str());
      
		    buffer=string("PID=") + string(pid) 
			 + ", installing call to DYNINSTinit()";
		    statusLine(buffer.c_str());
		} else {
			   
			//breakpoint after attach....
#if !defined(BPATCH_LIBRARY) //ccw 6 jun 2002 SPLIT
			p->status_ = stopped;
	    		p->pause_();
		    	p->procStopFromDYNINSTinit();

		    	p->mungeAddr = loadParadynDll(p); 
			p->getDefaultLWP()->changePC(p->mungeAddr, NULL);
			p->isLoadingParadynLib = true;
 		        p->status_ = running;
	        	p->continueProc_();
#else
			p->pause_();
			p->handleCompletionOfDYNINSTinit(true); //this line may be wrong...
#endif
			break;
		}
	    }

	    // the process has already been initialized
	    // First lookup in the tramp table, to see if the breakpoint
	    // is from an instrumentation point.

	    // lookup in tramp table

#if !defined( mips_unknown_ce2_11 )//ccw 26 july 2000 : 29 mar 2001
	    Address trampAddr = 0;
	    unsigned u; unsigned k;
	    unsigned key = 
		(unsigned)debugEv.u.Exception.ExceptionRecord.ExceptionAddress;
	    for (u = HASH1(key); 1; u = (u + HASH2(key)) % TRAMPTABLESZ) {
		k = p->trampTable[u].key;
		if (k == 0)
		    break;
		else if (k == key) {
		    trampAddr = p->trampTable[u].val;
		    break;
		}
	    }
	    if (trampAddr) {
		    // this is a trap from an instrumentation point

	      // find the current thread
	      dyn_thread* currThread = NULL;
	      for(unsigned int i = 0; i < p->threads.size(); i++)
		{
		  if ((unsigned)p->threads[i]->get_tid() == debugEv.dwThreadId)
		    {
		      currThread = p->threads[i];
		      break;
		    }
		}
	      assert( currThread != NULL );
	      
	      // Due to a race between the processing of trap debug events
	      // and the desire to run an inferior RPC, it is possible that
	      // we hit the trap, set ourselves up to run an inferior RPC,
	      // and then processed the trap notification.  If this happens,
	      // we don't end up running *any* of the inferior RPC code.
	      //
	      // We can tell that this is what's happened based on the
	      // thread's Eip - if it doesn't match the ExceptionAddress, we
	      // know that we had tried to reset the Eip to execute an inferior
	      // RPC.  In that case, we leave the Eip alone here, which means
	      // the inferior RPC code will execute when we continue the 
	      // thread.
	      // We have to remember that we need to execute this 
	      // instrumentation once the inferior RPC is done, however.

	    CONTEXT cont;
	    cont.ContextFlags = CONTEXT_FULL;
	    if( !GetThreadContext( (HANDLE)currThread->get_lwp()->get_fd(), &cont ) )
	      assert(0 && "Failed to get thread context");
            if( debugEv.u.Exception.ExceptionRecord.ExceptionAddress == (PVOID)(cont.Eip - 1) )
	      {
                // The Eip indicates we've just executed a trap instruction
                // reset the Eip to the address of the base tramp
                cont.Eip = trampAddr;
                if( !SetThreadContext( (HANDLE)currThread->get_lwp()->get_fd(), &cont ) )
		  assert(0 && "Failed to set thread context");
	      }
            else {
	      // We leave the Eip alone, since we've set it to execute 
	      // at inferior RPC code.  However, we need to remember that
	      // we should execute the base tramp once the inferiorRPC is
	      // completed.
	      currThread->set_pending_tramp_addr( trampAddr );
            }
	    
	    break;
	    }
#endif // mips_unknown_ce2_11
	    
	    // If it is not from an instrumentation point,
	    // it could be from a call to DYNINSTbreakPoint
	    // or an inferior procedure call

	    p->status_ = stopped;
	    p->pause_();
	    int result = p->procStopFromDYNINSTinit();
	    assert(result >= 0 && result <= 2);
	    if (result != 0) {
#if !defined(BPATCH_LIBRARY) //ccw 5 jun 2002 SPLIT
	    	p->mungeAddr = loadParadynDll(p); 
		p->getDefaultLWP()->changePC(p->mungeAddr, NULL);
		p->isLoadingParadynLib = true;
 	        p->status_ = running;
	        p->continueProc_();
#endif
		if (result == 1) {
#if defined( mips_unknown_ce2_11 )//ccw 20 mar 2001 : 29 mar 2001
			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 1 mar 2001
#endif			
	    		//p->pause_();
		} else {
		    // procStopFromDYNINSTinit called continueProc()
		    // What should we do here?
		    
		    break;
		}
	    } else if (p->handleTrapIfDueToRPC()) {
		//p->pause_();
		break;
	    } else {
#ifdef BPATCH_LIBRARY /* Don't ignore unknown bkpt in library; leave stopped. */
		*status = SIGEM_STOPPED | SIGTRAP;
#if defined( mips_unknown_ce2_11 )//ccw 20 mar 2001 : 29 mar 2001
			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 1 mar 2001
#endif

#else
		// Unknown breakpoint: we just ignore it
		printf("Debug breakpoint exception, %d, addr = %x\n", 
		       debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
		       debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
		p->status_ = running;
		p->continueProc_();
#endif
	    }
	    break;
	}
	case EXCEPTION_ILLEGAL_INSTRUCTION:
	  //printf("Illegal instruction\n");
	  p->pause_();
	  p->status_ = stopped;
	  if( p->handleTrapIfDueToRPC() )
	    {
	      // handleTrapIfDueToRPC calls continueProc()
	      // however, under Windows NT, it doesn't actually 
	      // continue the thread until the ContinueDbgEvent call is made
	      
	      // We take advantage of this fact to ensure that any
	      // pending instrumentation is executed.  (I.e., instrumentation
	      // put off so we could be sure to execute inferior RPC code is
	      // now executed.)
	      if( !p->isRunningIRPC() )	{
		// we finished the inferior RPC, so we can now execute
		// any pending instrumentation
		
		// find the current thread
		dyn_thread* currThread = NULL;
		for( unsigned int i = 0; i < p->threads.size(); i++ ) {
		  if ((unsigned)p->threads[i]->get_tid() == 
		      debugEv.dwThreadId) {
		    currThread = p->threads[i];
		    break;
		  }
		}
		assert( currThread != NULL );
		
		Address pendingTrampAddr = currThread->get_pending_tramp_addr();
		if( pendingTrampAddr) {
		  // reset the Eip to the pending instrumentation
		  CONTEXT ctxt;
		  ctxt.ContextFlags = CONTEXT_FULL;
		  if( !GetThreadContext( (HANDLE)currThread->get_lwp()->get_fd(), &ctxt ) )
		    assert(0 && "Failed to get thread context");
		  ctxt.Eip = pendingTrampAddr;
		  if( !SetThreadContext( (HANDLE)currThread->get_lwp()->get_fd(), &ctxt ) )
		    assert(0 && "Failed to set thread context");
		  currThread->set_pending_tramp_addr(0);
                }
	      }
	      break;
	    }
	  p->status_ = running;
	  p->continueProc_();
#if defined( mips_unknown_ce2_11) //ccw 28 july 2000 : 29 mar 2001
	  BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
							    DBG_EXCEPTION_NOT_HANDLED);
#else
	  ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			     DBG_EXCEPTION_NOT_HANDLED);
#endif
	  return 0;
	  
	case EXCEPTION_ACCESS_VIOLATION:
	  printf("Access violation exception, %d, addr = %08x\n", 
		 debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
		 debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
	  dumpMem(p, debugEv.u.Exception.ExceptionRecord.ExceptionAddress, 32);
	  
        {

#ifndef mips_unknown_ce2_11 //ccw 6 feb 2001 : 29 mar 2001
	  // Should walk stacks for other threads as well
	  pdvector<pdvector<Frame> > stackWalks;
      p->walkStacks(stackWalks);
      for (unsigned walk_iter = 0; walk_iter < stackWalks.size(); walk_iter++)
          for( unsigned i = 0; i < stackWalks[walk_iter].size(); i++ )
          {
              function_base* f = p->findFuncByAddr( stackWalks[walk_iter][i].getPC() );
              const char* szFuncName = (f != NULL) ? f->prettyName().c_str() : "<unknown>";
              fprintf( stderr, "%08x: %s\n", stackWalks[walk_iter][i].getPC(), szFuncName );
          }
#endif
      
        }
	    //	ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
	    //			   DBG_EXCEPTION_NOT_HANDLED);
	    //	break;
	default:
	    //printf("exeption %x, addr %x\n", exnCode, 
	    //   debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
#if defined( mips_unknown_ce2_11 )//ccw 28 july 2000 : 29 mar 2001
		BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
				   DBG_EXCEPTION_NOT_HANDLED);
#else
		ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
				   DBG_EXCEPTION_NOT_HANDLED);
#endif

	    return 0;
	}
    } break;

    case CREATE_THREAD_DEBUG_EVENT: {
	//printf("create thread, tid = %d\n", debugEv.dwThreadId);
      assert(p->threads.size() > 0); // main thread should be already defined
      dyn_lwp *l = new dyn_lwp(debugEv.dwThreadId, // ID
			       debugEv.u.CreateThread.hThread, // fd
			       p); // process
      l->openFD();
      dyn_thread *t = new dyn_thread(p, debugEv.dwThreadId, // thread ID
				 p->threads.size()-1, // POS (unused currently)
				 l); // dyn_lwp object for thread handle
      p->threads.push_back(t);
      p->lwps[debugEv.dwThreadId] = l; // Store in table of lwps
    } break;
    
    case CREATE_PROCESS_DEBUG_EVENT: {
	//CREATE_PROCESS_DEBUG_INFO info = debugEv.u.CreateProcessInfo;
	printf("CREATE_PROCESS event: %d\n", debugEv.dwProcessId);
	p = findProcess(debugEv.dwProcessId);
	if (p) {
	    p->secondBkpt = 0; //ccw 27 june 2001 : allows for more than one mutatee to be started
	    //fprintf(stderr,"create process: base = %x\n", info.lpBaseOfImage);
        dyn_lwp *l = p->getDefaultLWP();
        if (!l) {
            // It's possible we never created the default LWP
            l = new dyn_lwp(debugEv.dwThreadId,
                            debugEv.u.CreateProcessInfo.hThread, 
                            p);
            p->lwps[0] = l;
        }
	    if (p->threads.size() == 0) {
            dyn_thread *t = new dyn_thread(p, debugEv.dwThreadId, // thread ID,
                                           0, // POS (main thread is always 0)
                                           l);
            // define the main thread
            p->threads.push_back(t);
        }
        else {
            p->threads[0]->update_tid(debugEv.dwThreadId);
            p->threads[0]->update_lwp(l);
        }
    }
    break;
    }
  case EXIT_THREAD_DEBUG_EVENT: {
	//printf("exit thread, tid = %d\n", debugEv.dwThreadId);
	unsigned nThreads = p->threads.size();
	// start from one to skip main thread
	for (unsigned u = 1; u < nThreads; u++) {
	    if ((unsigned)p->threads[u]->get_tid() == debugEv.dwThreadId) {
		delete p->threads[u];
		p->threads[u] = p->threads[nThreads-1];
		p->threads.resize(nThreads-1);
		break;
	    }
	}
    } break;

    case EXIT_PROCESS_DEBUG_EVENT:
	p = findProcess(debugEv.dwProcessId);
	if (p) {
            char errorLine[1024];
            sprintf(errorLine, "Process %d has terminated with code 0x%x\n", 
                p->getPid(), debugEv.u.ExitProcess.dwExitCode);
            statusLine(errorLine);
            logLine(errorLine);
	    handleProcessExit(p, debugEv.u.ExitProcess.dwExitCode);
	}
#ifdef BPATCH_LIBRARY
	*status = SIGEM_EXITED;
#else
	break;
#endif
	break; //ccw 23 july 2001
    case LOAD_DLL_DEBUG_EVENT: {
	//printf("load dll: hFile=%x, base=%x, debugOff=%x, debugSz=%d lpname=%x, %d\n",
	//       debugEv.u.LoadDll.hFile, debugEv.u.LoadDll.lpBaseOfDll,
	//       debugEv.u.LoadDll.dwDebugInfoFileOffset,
	//       debugEv.u.LoadDll.nDebugInfoSize,
	//       debugEv.u.LoadDll.lpImageName,
	//       /*debugEv.u.LoadDll.fUnicode*/
	//       GetFileSize(debugEv.u.LoadDll.hFile,NULL));

	// We are currently not handling dynamic libraries loaded after the
	// application starts.
	if (p->isBootstrappedYet()
#if !defined(BPATCH_LIBRARY) //ccw 4 jun 2002 SPLIT
			&& p->isPARADYNBootstrappedYet()
#endif
			)
	  break;


	// obtain the name of the DLL
	string imageName = GetLoadedDllImageName( p, debugEv );


	// try to load symbols for the DLL
        if (!SymLoadModule((HANDLE)p->getProcessHandle(),
			   debugEv.u.LoadDll.hFile,
			   NULL, NULL, 0, 0)) {

		char msgText[1024];

		sprintf( msgText, "SymLoadModule failed for %s: 0x%x\n",
			imageName.c_str(), GetLastError() );

		logLine(msgText);
	}

	// discover structure of new DLL, and incorporate into our
	// list of known DLLs
	if (imageName.length() > 0) {
	shared_object *so = 
#if defined( mips_unknown_ce2_11 )//ccw 29 mar 2001
	new shared_object(imageName, (unsigned long) debugEv.u.LoadDll.lpBaseOfDll,
		false, true, true, 0);
#else
		new shared_object(imageName, 0,false, true, true, 0);
#endif
	    assert(p->dyn);
	    p->dyn->sharedObjects.push_back(so);
	    if (!p->shared_objects) {
	      p->shared_objects = new pdvector<shared_object *>;
	    }
	    (*(p->shared_objects)).push_back(so);
#ifndef BPATCH_LIBRARY
	    tp->resourceBatchMode(true);
#endif 
	    p->addASharedObject(*so,(Address) debugEv.u.LoadDll.lpBaseOfDll); //ccw 20 jun 2002
#ifndef BPATCH_LIBRARY
	    tp->resourceBatchMode(false);
#endif 
	    p->setDynamicLinking();

//#ifdef BPATCH_LIBRARY
            {
                char dllFilename[_MAX_FNAME];
                char dllExtension[_MAX_EXT];

                _splitpath( imageName.c_str(),
                            NULL,
                            NULL,
                            dllFilename,
                            dllExtension );
                if( !_stricmp( dllFilename, "libdyninstAPI_RT" ) &&
                    !_stricmp( dllExtension, ".dll" ) ) {
                    // ccw 2 may 2001
                    //when we load libdyninstAPI_RT.dll we need to find the local
                    //variables in the dll that will hold the cause and pid so when
                    //DLLMain is called they will be correctly passed to DYNINSTinit
		    //printf(" WRITING VALUE TO DYNINT LIB\n");//PRINTF
                    if(!setVariable(p, "libdyninstAPI_RT_DLL_localCause", 1)){
                         assert(0);
                    }
                    if(!setVariable(p, "libdyninstAPI_RT_DLL_localPid", getpid())){
                         assert(0);
                    }
                }
            }
//#else
#if !defined(BPATCH_LIBRARY)  //ccw 4 jun 2002 SPLIT

	    {

                char dllFilename[_MAX_FNAME];
                char dllExtension[_MAX_EXT];

                _splitpath( imageName.c_str(),
                            NULL,
                            NULL,
                            dllFilename,
                            dllExtension );
                if( !_stricmp( dllFilename, "libdyninstRT" ) &&
                    !_stricmp( dllExtension, ".dll" ) ) {
                    // ccw 2 may 2001
                    //when we load libdyninstRT.dll we need to find the local
                    //variables in the dll that will hold the cause and pid so when
                    //DLLMain is called they will be correctly passed to pDYNINSTinit
		 	// variable 1
			int shmKey;
#ifdef SHM_SAMPLING
			shmKey = (p->getShmKeyUsed());
#else 
			shmKey = 0;
#endif
			// variable 2
			int shmSize;
#ifdef SHM_SAMPLING
			shmSize = (p->getShmHeapTotalNumBytes());
#else 
			shmSize = 0;
#endif
			// variable 3
			int pPid;
			if(p->wasCreatedViaAttach() ){
				pPid = (-1 * traceConnectInfo);
			}else{
				pPid = (1 * traceConnectInfo);
			}

			int numThreads = p->maxNumberOfThreads();

			Address offsetOfShMetaOffsetData = 
			   p->initSharedMetaData();

			//printf(" SETTING VALUES IN PARADYN LIB %d %d %d\n",var1, var2, var3);//PRINTF
			//fflush(stdout);
                    if(!setVariable(p, "libdyninstRT_DLL_localtheKey", shmKey)){
                         assert(0);
                    }
                    if(!setVariable(p, "libdyninstRT_DLL_localshmSegNumBytes", shmSize)){
                         assert(0);
                    }
		    if(!setVariable(p, "libdyninstRT_DLL_localparadynPid", pPid)){
                         assert(0);
                    }
		    if(!setVariable(p, "libdyninstRT_DLL_localmaxThreads", numThreads)) {
		      assert(0);
		    }
		    if(!setVariable(p, "libdyninstRT_DLL_localoffset", 
				    offsetOfShMetaOffsetData)) {
		      assert(0);
		    }
                }
	    }
#endif	
	  }
		//ccw 31 jan 2001 : 29 mar 2001
		//here i need to check to see if this is the first breakpt since
		//creating the process and if it is, do whatever should happen when 
		//i get the second NT breakpoint.
		//This is a HACK that fixes the bug that WinCE does NOT send a
		//debug message before starting main, you get one when you create
		//the process (before loading and dlls and any initializations take place)
		//and that is it.  so i grab one after coredll.dll is loaded
		//and load the libdyninstAPI_RT.dll

#ifdef mips_unknown_ce2_11 //ccw 31 jan 2001 : 29 mar 2001
		if(!_stricmp(imageName.c_str(),"coredll.dll")){ 
			//make sure we have loaded coredll.dll --this contains LoadLibrary

			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 25 oct 2000
			if (!p->reachedFirstBreak) {

				if (!p->hasLoadedDyninstLib && !p->isLoadingDyninstLib) {
					//DebugBreak();//ccw 30 jan 2001
					continueType = DBG_CONTINUE;//ccw 25 oct 2000
					//DebugBreak();
					Address addr = loadDyninstDll(p, p->savedData);
					p->savedRegs = p->getDefaultLWP()->getRegisters();
					p->getDefaultLWP()->changePC(addr, NULL);
					//p->LoadDyninstTrapAddr = addr + 0xd;
					p->isLoadingDyninstLib = true;
					break;
				}
			}
		}
#endif
	
	}
    break;
    case UNLOAD_DLL_DEBUG_EVENT:
	// TODO
	//printf("unload dll\n");
	break;
    case OUTPUT_DEBUG_STRING_EVENT:
#if defined( mips_unknown_ce2_11 ) //ccw 29 mar 2001
	//printf("output debug string\n");
		char dbgStr[512];//ccw 3 oct 2000
		memset(dbgStr,'\0', 512); //ccw 3 oct 2000
		p->readDataSpace_(debugEv.u.DebugString.lpDebugStringData, //ccw 3 oct 2000
			(debugEv.u.DebugString.nDebugStringLength<512) ? 
			debugEv.u.DebugString.nDebugStringLength:511 , dbgStr );
		wprintf(L"output debug string: %s\n",dbgStr); //ccw 3 oct 2000
		//DebugBreak();
#endif
	//printf("output debug string\n");
	break;
    case RIP_EVENT:
	//printf("rip event\n");
	break;
    default:
	;
	//printf("unknown debug event\n");
	
    }
#if defined( mips_unknown_ce2_11 )//ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			    continueType)) {
#else
    if (!ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			    DBG_CONTINUE)) {
#endif
	DebugBreak();
	printf("ContinueDebugEvent failed\n");
	printSysError(GetLastError());
    }

#ifdef BPATCH_LIBRARY
    if (*status)
	return debugEv.dwProcessId;
    else
	return 0;
#else
    return 0;
#endif
}


// already setup on this FD.
// disconnect from controlling terminal 
void OS::osDisconnect(void) {
#ifdef notdef
  int ttyfd = open ("/dev/tty", O_RDONLY);
  ioctl (ttyfd, TIOCNOTTY, NULL); 
  P_close (ttyfd);
#endif
}


bool process::attach() {
  if (createdViaAttach) {
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteDebugActiveProcess(getPid()))
#else
    if (!DebugActiveProcess(getPid()))
#endif
      {
	//printf("Error: DebugActiveProcess failed\n");
	return false;
      }
  }

#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
  procHandle_ = BPatch::bpatch->rDevice->RemoteOpenProcess(PROCESS_ALL_ACCESS, false, getPid());
#else
  procHandle_ = OpenProcess(PROCESS_ALL_ACCESS, false, getPid());
#endif

  
  if (procHandle_ == NULL) {
    //printf("Error: OpenProcess failed\n");
    assert(0);
  }
  
  void initSymbols(HANDLE procH, const string file, const string dir);
  initSymbols((HANDLE)procHandle_, symbols->file(), "");
  if (createdViaAttach) {
    //
    //void initSymbols(HANDLE procH, const string file, const string dir);
    //initSymbols((HANDLE)getDefaultLWP()->get_fd(), symbols->file(), "");
  }
  return true;
}

/* continue a process that is stopped */
bool process::continueProc_() {
  for (unsigned u = 0; u < threads.size(); u++) {
    unsigned count;

#ifdef mips_unknown_ce2_11 //ccw 10 feb 2001 : 29 mar 2001
    count = BPatch::bpatch->rDevice->RemoteResumeThread((HANDLE)threads[u]->get_lwp()->get_fd());
#else
    count = ResumeThread((HANDLE)threads[u]->get_lwp()->get_fd());
#endif
    if (count == 0xFFFFFFFF) {
      printSysError(GetLastError());
      return false;
    }
  }
  return true;
}


#ifdef BPATCH_LIBRARY
/*
   terminate execution of a process
 */
bool process::terminateProc_()
{
    OS::osKill(pid);
    handleProcessExit(this, -1);
    return true;
}
#endif


/*
   pause a process that is running
*/
bool process::pause_() {
    for (unsigned u = 0; u < threads.size(); u++) {
	unsigned count ;
#ifdef mips_unknown_ce2_11 //ccw 10 feb 2001 : 29 mar 2001
	count = BPatch::bpatch->rDevice->RemoteSuspendThread((HANDLE)threads[u]->get_lwp()->get_fd());
#else
	count = SuspendThread((HANDLE)threads[u]->get_lwp()->get_fd());
#endif

	if (count == 0xFFFFFFFF) {
	    // printf("pause_: %d\n", threads[u]->get_tid());
	    // printSysError(GetLastError());
	    return false;
	} 
    }
    return true;
}

/*
   close the file descriptor for the file associated with a process
*/
bool process::detach_() {
    return false;
}


#ifdef BPATCH_LIBRARY
/*
   detach from thr process, continuing its execution if the parameter "cont"
   is true.
 */
bool process::API_detach_(const bool cont)
{
    // XXX Not yet implemented
    assert(false);
    return false;
}
#endif


bool process::dumpCore_(const string) {
    return false;
}

bool process::writeTextWord_(caddr_t inTraced, int data) {
    return writeDataSpace_(inTraced, sizeof(int), (caddr_t) &data);
}

bool process::writeTextSpace_(void *inTraced, u_int amount, const void *inSelf) {
    return writeDataSpace_(inTraced, amount, inSelf);
}

bool process::flushInstructionCache_(void *baseAddr, size_t size){ //ccw 25 june 2001
	return FlushInstructionCache((HANDLE)procHandle_, baseAddr, size);
}

#ifdef BPATCH_SET_MUTATIONS_ACTIVE
bool process::readTextSpace_(void *inTraced, u_int amount, const void *inSelf) {
  return readDataSpace_(inTraced, amount, (void *)inSelf);
}
#endif

bool process::writeDataSpace_(void *inTraced, u_int amount, const void *inSelf) {
    DWORD nbytes;

    //printf("write %d bytes, %x\n", amount, inTraced);
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    bool res = BPatch::bpatch->rDevice->RemoteWriteProcessMemory((HANDLE)procHandle_, (LPVOID)inTraced, 
				  (LPVOID)inSelf, (DWORD)amount, &nbytes);
#else
    bool res = WriteProcessMemory((HANDLE)procHandle_, (LPVOID)inTraced, 
				  (LPVOID)inSelf, (DWORD)amount, &nbytes);
#endif
    return res && (nbytes == amount);
}


bool process::readDataSpace_(const void *inTraced, u_int amount, void *inSelf) {
    DWORD nbytes;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    bool res = BPatch::bpatch->rDevice->RemoteReadProcessMemory((HANDLE)procHandle_, (LPVOID)inTraced, 
				 (LPVOID)inSelf, (DWORD)amount, &nbytes);
#else
    bool res = ReadProcessMemory((HANDLE)procHandle_, (LPVOID)inTraced, 
				 (LPVOID)inSelf, (DWORD)amount, &nbytes);
#endif
    return res && (nbytes == amount);
}


bool process::loopUntilStopped() { assert(0); return false; }

float OS::compute_rusage_cpu() { return 0.0; }
float OS::compute_rusage_sys() { return 0.0; }
float OS::compute_rusage_min() { return 0.0; }
float OS::compute_rusage_maj() { return 0.0; }
float OS::compute_rusage_swap() { return 0.0; }
float OS::compute_rusage_io_in() { return 0.0; }
float OS::compute_rusage_io_out() { return 0.0; }
float OS::compute_rusage_msg_send() { return 0.0; }
float OS::compute_rusage_msg_recv() { return 0.0; }
float OS::compute_rusage_sigs() { return 0.0; }
float OS::compute_rusage_vol_cs() { return 0.0; }
float OS::compute_rusage_inv_cs() { return 0.0; }


int getNumberOfCPUs() {
    SYSTEM_INFO info;
    GetSystemInfo(&info);
    return info.dwNumberOfProcessors;
}  


Frame dyn_lwp::getActiveFrame()
{
  w32CONTEXT cont; //ccw 27 july 2000 : 29 mar 2001
  
  Address pc = 0, fp = 0, sp = 0;
  
  // we must set ContextFlags to indicate the registers we want returned,
  // in this case, the control registers.
  // The values for ContextFlags are defined in winnt.h
  cont.ContextFlags = CONTEXT_CONTROL;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
  if (BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)get_fd(), &cont))
#else
  if (GetThreadContext((HANDLE)get_fd(), &cont))
#endif
    {
#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
      fp = cont.Ebp;
      pc = cont.Eip;
#elif mips_unknown_ce2_11
      pc = cont.Fir;
      sp = cont.IntSp;
      fp = cont.IntS8;
#endif
      return Frame(pc, fp, sp, proc_->getPid(), NULL, this, true);
    }
    printSysError(GetLastError());
    return Frame();
}

#if defined(i386_unknown_nt4_0) //ccw 29 mar 2001
 

Frame Frame::getCallerFrame(process *p) const
{
    // For x86, the frame-pointer (EBP) points to the previous 
    // frame-pointer, and the saved return address is in EBP-4.

    struct {
	Address fp;
	Address rtn;
    } addrs;

    if (p->readDataSpace((caddr_t)(fp_), sizeof(int)*2,
			 (caddr_t)&addrs, true))
    {
        Frame ret;
        ret.fp_ = addrs.fp;
	ret.pc_ = addrs.rtn;
	return ret;
    }

    return Frame(); // zero frame
}

#elif mips_unknown_ce2_11 //ccw 6 feb 2001

Frame Frame::getCallerFrame(process *p) const
{
//	DebugBreak();//ccw 6 feb 2001

	Frame ret;
	Address prevPC;

	//find which function we are in to determine the frame size

	// (i.e. $pc in native/basetramp/minitramp code)
	instPoint     *ip = NULL;
	trampTemplate *bt = NULL;
	instInstance  *mt = NULL;
	pd_Function *callee = findAddressInFuncsAndTramps(p, pc_, ip, bt, mt);
	// non-NULL "ip" means that $pc is in instrumentation
	// non-NULL "bt" means that $pc is in basetramp
	// non-NULL "mt" means that $pc is in minitramp
	if (ip) assert(bt || mt);

	ret.sp_ = sp_ - callee->frame_size;

	Address tmpSp = sp_ + 20;
	p->readDataSpace((caddr_t)(tmpSp), sizeof(int),
			 &prevPC, true);

	ret.pc_ = prevPC;
	  

	return ret;
}
#endif

struct dyn_saved_regs *dyn_lwp::getRegisters() {
    struct dyn_saved_regs *regs = new dyn_saved_regs();
    
  // we must set ContextFlags to indicate the registers we want returned,
  // in this case, the control registers.
  // The values for ContextFlags are defined in winnt.h
  regs->cont.ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000 : 29 mar 2001
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
  if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)get_fd(), &(regs->cont)))
#else
  handleT handle = get_fd();
  if (!GetThreadContext((HANDLE)handle, &(regs->cont)))
#endif
    {
      return NULL;
    }
  return regs;
}

bool dyn_lwp::changePC(Address addr, struct dyn_saved_regs *regs)
{
  w32CONTEXT cont;//ccw 27 july 2000
  if (!regs) {
      cont.ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000 : 29 mar 2001
      //	DebugBreak();
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
      if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)get_fd(), &cont)) 
#else
          if (!GetThreadContext((HANDLE)get_fd(), &cont))
#endif
          {
              printf("GetThreadContext failed\n");
              return false;
          }
  }
  else {
      memcpy(&cont, &(regs->cont), sizeof(w32CONTEXT));
  }
#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001 
  cont.Eip = addr;
#elif  mips_unknown_ce2_11 
  cont.Fir = addr;
#endif
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
  if (!BPatch::bpatch->rDevice->RemoteSetThreadContext((HANDLE)get_fd(), &cont))
#else
  if (!SetThreadContext((HANDLE)get_fd(), &cont))
#endif
    {
      printf("SethreadContext failed\n");
      return false;
    }
  return true;
}

bool dyn_lwp::restoreRegisters(struct dyn_saved_regs *regs) {
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
  if (!BPatch::bpatch->rDevice->RemoteSetThreadContext((HANDLE)get_fd(),
                                                       &(regs->cont)))
#else
  if (!SetThreadContext((HANDLE)get_fd(), 
                        &(regs->cont)))
#endif
    {
      //printf("SetThreadContext failed\n");
      return false;
    }
  return true;
}

bool process::isRunning_() const {
    // TODO
    printf("process::isRunning_() returning true\n");
    return true;
}


string process::tryToFindExecutable(const string& iprogpath, int pid) {
  return iprogpath;
}

bool process::set_breakpoint_for_syscall_completion() {
   /* Can assume: (1) process is paused and (2) in a system call.
      We want to set a TRAP for the syscall exit, and do the
      inferiorRPC at that time.  We'll use /proc PIOCSEXIT.
      Returns true iff breakpoint was successfully set. */

    // This is never called on Windows NT
    assert(false);
    return false;
}

bool process::clear_breakpoint_for_syscall_completion() { return true; }

Address dyn_lwp::readRegister(Register reg)
{
   w32CONTEXT *cont = new w32CONTEXT;//ccw 27 july 2000 : 29 mar 2001
    if (!cont)
	return NULL;
    // we must set ContextFlags to indicate the registers we want returned,
    // in this case, the control registers.
    // The values for ContextFlags are defined in winnt.h
    cont->ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)get_fd(), cont)) {
#else
    if (!GetThreadContext((HANDLE)get_fd(), cont)) {
#endif
	delete cont;
	return NULL;
    }
#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
    return cont->Eax;
#elif mips_unknown_ce2_11 
	return (&cont->IntZero)[reg]; 
	//the registers in the MIPS context returned by
	//winCE return general registers 0..31 in the struct
	//starting at IntZero and going to IntRa, each being a DWORD
#endif
}

bool dyn_lwp::executingSystemCall() {
   // TODO
   return false;
}


void initSymbols(HANDLE procH, const string file, const string dir) {
  string searchPath;
  char sysRootDir[MAX_PATH+1];
  if (GetSystemDirectory(sysRootDir, MAX_PATH) == 0)
     assert(0);
  string sysSymsDir = string(sysRootDir) + "\\..\\symbols";
  if (dir.length())
    searchPath = dir + ";";
  searchPath = searchPath + sysSymsDir + ";" + sysSymsDir + "\\dll";
  if (!SymInitialize(procH, (char *)searchPath.c_str(), 0)) {
    fprintf(stderr,"SymInitialize failed, %x\n", GetLastError()); fflush(stderr);
    return;
  }
  if (!SymLoadModule(procH, NULL, (char *)file.c_str(), NULL, 0, 0)) {
    printf("SymLoadModule failed for \"%s\", %x\n",
	    file.c_str(), GetLastError());
    return;
  }
}





/*****************************************************************************
 * forkNewProcess: starts a new process, setting up trace and io links between
 *                the new process and the daemon
 * Returns true if succesfull.
 * 
 * Arguments:
 *   file: file to execute
 *   dir: working directory for the new process
 *   argv: arguments to new process
 *   envp: environment **** not in use
 *   inputFile: where to redirect standard input
 *   outputFile: where to redirect standard output
 *   traceLink: handle or file descriptor of trace link (read only)
 //removed all ioLink related code for output redirection
 *   ioLink: handle or file descriptor of io link (read only)
 *   pid: process id of new process
 *   tid: thread id for main thread (needed by WindowsNT)
 *   procHandle: handle for new process (needed by WindowsNT)
 *   thrHandle: handle for main thread (needed by WindowsNT)
 ****************************************************************************/
bool forkNewProcess(string &file, string dir, pdvector<string> argv, 
		    pdvector<string>envp, string inputFile, string outputFile,
		    int &traceLink,  
		    int &pid, int &tid, 
		    int &procHandle, int &thrHandle, int /* stdin_fd */, 
		    int stdout_fd, int /* stderr_fd */) {
#ifdef mips_unknown_ce2_11 //ccw 8 aug 2000 : 29 mar 2001
	WCHAR appName[256];
	WCHAR dirName[256];
	WCHAR commLine[256];
#endif

#ifndef BPATCH_LIBRARY
#ifdef notdef_Pipes     // isn't this all obsolete???
    HANDLE rTracePipe;
    HANDLE wTracePipe;
    // security attributes to make handles inherited
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = true;
    sa.lpSecurityDescriptor = NULL;
    
    // create trace pipe
    if (!CreatePipe(&rTracePipe, &wTracePipe, &sa, 0)) {
	string msg = string("Unable to create trace pipe for program '") 
	    + File + string("': ") + string(sys_errlist[errno]);
	showErrorCallback(68, msg);
	return(NULL);
    }
    
    /* removed for output redirection
    HANDLE rIoPipe;
    HANDLE wIoPipe;
    // create IO pipe
    // ioPipe is used to redirect the child's stdout & stderr to a pipe which
    // is in turn read by the parent via the process->ioLink socket.
    if (!CreatePipe(&rIoPipe, &wIoPipe, &sa, 0)) {
	string msg = string("Unable to create IO pipe for program '") + File +
	    string("': ") + string(sys_errlist[errno]);
	showErrorCallback(68, msg);
	return(NULL);
    }
    SetEnvironmentVariable("PARADYN_IO_PIPE",
			   string((unsigned)wIoPipe).c_str());
    */

    printf("tracepipe = %d\n", (unsigned)wTracePipe);
    // enter trace and IO pipes in child's environment
    SetEnvironmentVariable("PARADYN_TRACE_PIPE", 
			   string((unsigned)wTracePipe).c_str());
#endif
    //  extern int traceSocket;
    //  SetEnvironmentVariable("PARADYND_TRACE_SOCKET", string((unsigned)traceSocket).c_str());
#endif /* BPATCH_LIBRARY */
    
    // create the child process
    
    string args;
    for (unsigned ai=0; ai<argv.size(); ai++) {
	args += argv[ai];
	args += " ";
    }
    
    STARTUPINFO stinfo;
    memset(&stinfo, 0, sizeof(STARTUPINFO));
    stinfo.cb = sizeof(STARTUPINFO);

    /*to do: output redirection
    //stinfo.hStdOutput = (HANDLE)ioLink;
    stinfo.hStdOutput = (HANDLE)stdout_fd;
    stinfo.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
    stinfo.hStdError = GetStdHandle(STD_ERROR_HANDLE);
    stinfo.dwFlags |= STARTF_USESTDHANDLES;
    */
    PROCESS_INFORMATION procInfo;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
	kludgeWCHAR(file.c_str(),appName);
	kludgeWCHAR(args.c_str(),commLine);
	if(dir == ""){
		dirName[0] = (unsigned short) 0;
	}else{
		kludgeWCHAR(dir.c_str(),dirName);
	}

	/* ccw 8 aug 2000  : 29 mar 2001
	 WARNING:  
		If compiler optimization are turned on (-Ox) for remoteDevice.C the following
		line of code bugs. An extra parameter is passed to the function before 
		appName and consequently each argument is shifted one to the right, 
		completely destroying the code.
		
		DO NOT TURN ON -Ox for remoteDevice.C!

		if need be, pdwinnt.C can be compiled with -Ox
		less optimization may be safe, it has not yet been tried. if someone is bored
		they should try different optimization flags to see if any work correctly here.
		Until then the mips/ce version of dyninst will be compiled w/o ANY optimizations!
	*/
	//DebugBreak();
    if (BPatch::bpatch->rDevice->RemoteCreateProcess(appName , commLine,
		      NULL, NULL, false,
		      (DWORD) DEBUG_PROCESS  /* | CREATE_NEW_CONSOLE /* | CREATE_SUSPENDED */,
		      NULL, dirName, 
		      &stinfo, &procInfo)) {

#else

    if (CreateProcess(file.c_str(), (char *)args.c_str(), 
		      NULL, NULL, TRUE,
		      DEBUG_PROCESS /* | CREATE_NEW_CONSOLE /* | CREATE_SUSPENDED */,
		      NULL, dir == "" ? NULL : dir.c_str(), 
		      &stinfo, &procInfo)) {
#endif

	procHandle = (Word)procInfo.hProcess;
	thrHandle = (Word)procInfo.hThread;
	pid = (Word)procInfo.dwProcessId;
	tid = (Word)procInfo.dwThreadId;
#ifdef notdef_Pipes
        traceLink = (Word)rTracePipe;
	CloseHandle(wTracePipe);
	/ *removed for output redirection
        //ioLink = (Word)rIoPipe;
        CloseHandle(wIoPipe);
        */
        //initSymbols((HANDLE)procHandle, file, dir);
#else
        traceLink = -1;
	/* removed for output redirection
        ioLink = -1;
	*/
#endif
	return true;
    }
   
#ifndef BPATCH_LIBRARY
#ifdef notdef_Pipes
    CloseHandle(rTracePipe);
    CloseHandle(wTracePipe);
    / *removed for output redirection
    CloseHandle(rIoPipe);
    CloseHandle(wIoPipe);
    */
#endif
#endif /* BPATCH_LIBRARY */

    // Output failure message
    LPVOID lpMsgBuf;

    if (FormatMessage( 
	    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
	    NULL,
	    GetLastError(),
	    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
	    (LPTSTR) &lpMsgBuf,
	    0,
	    NULL 
	) > 0) {
	char *errorLine = (char *)malloc(strlen((char *)lpMsgBuf) +
					 file.length() + 64);
	if (errorLine != NULL) {
	    sprintf(errorLine, "Unable to start %s: %s\n", file.c_str(),
		    (char *)lpMsgBuf);
	    logLine(errorLine);
	    showErrorCallback(68, (const char *) errorLine);

	    free(errorLine);
	}

	// Free the buffer returned by FormatMsg
	LocalFree(lpMsgBuf);
    } else {
	char errorLine[512];
	sprintf(errorLine, "Unable to start %s: unknown error\n",
		file.c_str());
	logLine(errorLine);
	showErrorCallback(68, (const char *) errorLine);
    }

    return false;
}

char *cplus_demangle(char *c, int) { 
    char buf[1000];
    if (c[0]=='_') {
        // VC++ 5.0 seems to decorate C symbols differently to C++ symbols
        // and the UnDecorateSymbolName() function provided by imagehlp.lib
        // doesn't manage (or want) to undecorate them, so it has to be done
        // manually, removing a leading underscore from functions & variables
        // and the trailing "$stuff" from variables (actually "$Sstuff")
        unsigned i;
        for (i=1; i<sizeof(buf) && c[i]!='$' && c[i]!='\0'; i++)
           { buf[i-1]=c[i]; }
        if (i==1) return 0; // avoid null names which seem to annoy Paradyn
        buf[i-1]='\0';
        return strdup(buf);
      }
    else if (UnDecorateSymbolName(c, buf, 1000, UNDNAME_NAME_ONLY)) {
	//printf("Undecorate: %s = %s\n", c, buf);

        // many symbols have a name like foo@4, we must remove the @4
        // just searching for an @ is not enough,
        // as it may occur on other positions. We search for the last one
        // and check that it is followed only by digits.
        char *p = strrchr(buf, '@');
	if (p) {
	  char *q = p+1;
	  strtoul(p+1, &q, 10);
	  if (q > p+1 && *q == '\0') {
	    *p = '\0';
	  }
	}
	return strdup(buf);
    }
    return 0;
}

bool OS::osKill(int pid) {
    bool res;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    HANDLE h = BPatch::bpatch->rDevice->RemoteOpenProcess(PROCESS_ALL_ACCESS, false, pid);
#else
    HANDLE h = OpenProcess(PROCESS_ALL_ACCESS, false, pid);
#endif
    if (h == NULL) {
	return false;
    }
#ifdef mips_unknown_ce2_11 //ccw 31 aug 2000 : 29 mar 2001
    res = BPatch::bpatch->rDevice->RemoteTerminateProcess(h,0);
    BPatch::bpatch->rDevice->RemoteCloseHandle(h);
#else
    res = TerminateProcess(h,0);
    CloseHandle(h);
#endif
    return res;
}

#if !defined(BPATCH_LIBRARY)
rawTime64 dyn_lwp::getRawCpuTime_hw() {
  fprintf(stderr, "dyn_lwp::getRawCpuTime_hw: not implemented\n");
  return 0;
}

rawTime64
FILETIME2rawTime64( FILETIME& ft )
{
    return (((rawTime64)(ft).dwHighDateTime<<32) | 
	    ((rawTime64)(ft).dwLowDateTime));
}

/* return unit: nsecs */
rawTime64 dyn_lwp::getRawCpuTime_sw() 
{
  FILETIME kernelT, userT, creatT, exitT;
  rawTime64 now;
  if(GetThreadTimes( (HANDLE)fd_,
      &creatT, &exitT, &kernelT,&userT)==0) {
    return 0;
  }

  // GetProcessTimes returns values in 100-ns units
  now = (FILETIME2rawTime64(userT)+FILETIME2rawTime64(kernelT));

  // time shouldn't go backwards, but we'd better handle it if it does
//  printf(" %I64d %I64d \n", now, previous);
  if (now < sw_previous_) {
	  //DebugBreak();
     logLine("********* time going backwards in paradynd **********\n");
     now = sw_previous_;
  }
  else {
     sw_previous_ = now;
  }

  return now;
}

bool process::catchupSideEffect(Frame &frame, instReqNode *inst)
{
  return true;
}

#endif

#ifdef mips_unknown_ce2_11 //ccw 2 aug 2000 : 29 mar 2001

#include "inst-mips.h"
//this comes from irix.C
bool process::heapIsOk(const pdvector<sym_data>&findUs)
{ 
	
  //ccw 12 oct 2000 NEED TO FIND WinMain here or _WinMain then try to find main or _main
//	DebugBreak();


  if (!(mainFunction = findOneFunction("WinMain")) &&
	  !(mainFunction = findOneFunction("_WinMain"))) {
		fprintf(stderr, "process::heapIsOk(): failed to find \"WinMain\"\n");

		if (!(mainFunction = findOneFunction("main")) &&
		!(mainFunction = findOneFunction("_main"))) {

			//if (!(mainFunction = findOneFunction("CFrameWnd::OnCreate")) &&
			//	!(mainFunction = findOneFunction("OnCreate"))) {

				//the following is for MFC apps
				//what is wrong with M$? all these different names for main()
				if (!(mainFunction = findOneFunction("wWinMain")) &&
					!(mainFunction = findOneFunction("_wWinMain"))) {


					fprintf(stderr, "process::heapIsOk(): failed to find \"main\"\n");
					return false;
				}
			//}
	  }	
  }

  for (unsigned i = 0; i < findUs.size(); i++) {
    const string &name = findUs[i].name;
    cerr << "Looking up symbol " << name << endl;
    /*
    Address addr = lookup_fn(this, name);
    if (!addr && findUs[i].must_find) {
      fprintf(stderr, "process::heapIsOk(): failed to find \"%s\"\n", name.c_str());
      return false;
    }
    */
  }

  return true;
}
#endif


fileDescriptor *getExecFileDescriptor(string filename, int &, bool)
{
  fileDescriptor *desc = new fileDescriptor(filename);
  return desc;
}

#ifndef BPATCH_LIBRARY
void process::initCpuTimeMgrPlt() {
  cpuTimeMgr->installLevel(cpuTimeMgr_t::LEVEL_TWO, &process::yesAvail, 
			   timeUnit(fraction(100)), timeBase::bNone(), 
			   &process::getRawCpuTime_sw, "swCpuTimeFPtrInfo");
}
#endif

bool getLWPIDs(pdvector <unsigned> &LWPids)
{
  assert (0 && "Not implemented");
  return false;
}




//
// This function retrieves the name of a DLL that has been
// loaded in an inferior process.  On the desktop flavors
// of Windows, the debug event that we get for loaded DLLs
// includes the location of a pointer to the DLL's image name.
// (Note the indirection.)  On Windows CE, the event contains
// the location of the image name string, with no indirection.
//
// There are several complications to overcome when reading this string:
//
// 1.  There is no guarantee that the image name is available.
//     In this case, the location in the debug event may be NULL,
//     or the pointer in the inferior process' address space may be NULL.
// 2.  The image name string may be either ASCII or Unicode.  Most of
//     the Windows system DLLs have Unicode strings, but many user-built
//     DLLs use single-byte strings.  If the string is Unicode, we need
//     to copy it to our address space and convert it to a single-byte
//     string because the rest of Paradyn/Dyninst has no clue what to
//     do with Unicode strings.
// 3.  We don't know how long the string is.  We have a loose upper
//     bound in that we know it is not more than MAX_PATH characters.
//     Unfortunately, the call we use to read from the inferior
//     process' address space will return failure if we ask for more
//     bytes than it can actually read (so we can't just issue a read
//     for MAX_PATH characters).  Given this limitation, we have to
//     try a read and check whether the read succeeded *and* whether
//     we read the entire image name string.  If not, we have to adjust
//     the amount we read and try again.
//
static
string
GetLoadedDllImageName( process* p, const DEBUG_EVENT& ev )
{
	string ret;


	if( ev.u.LoadDll.lpImageName == NULL )
	{
		// there is no image name string for us to read
		return ret;
	}

	char* msgText = new char[1024];	// buffer for error messages

#if defined(mips_unknown_ce2_11)

	// On Windows CE, the address given in the debug event struct
	// is the address of the DLL name string.
	void* pImageName = ev.u.LoadDll.lpImageName;
#else // defined(ce)
	// On non-CE flavors of Windows, the address given in the debug
	// event struct is the address of a pointer to the DLL name string.
	void* pImageName = NULL;
	if( !p->readDataSpace( ev.u.LoadDll.lpImageName, 4, &pImageName, false ) )
	{
		sprintf( msgText, "Failed to read DLL image name pointer: %d\n",
			GetLastError() );
		logLine( msgText );
	}
#endif // defined(ce)

	if( pImageName != NULL )
	{
		// we have the pointer to the DLL image name -
		// now read the name

		// allocate a conservatively-sized buffer
		char* buf = new char[(MAX_PATH + 1) * sizeof(WCHAR)];
		WCHAR* wbuf = (WCHAR*)buf;

		// We don't know how long the image name actually is, but
		// we do know that they tend not to be very long.
		// Therefore, we use a scheme to try to minimize the number
		// of reads needed to get the image name.
		// We do reads within ranges, starting with [1,128] bytes,
		// then [129,256] bytes, etc. up to MAX_PATH if necessary.
		// Within each range, we do reads following a binary search
		// algorithm.  For example, for the [1,128] range, we start
		// by trying to read 128 bytes.  If that read fails, we
		// try to half the number of bytes (i.e., 64).  If that
		// read also fails, we continue to halve the read requests 
		// until we find one that succeeds.
		//
		// When a read succeeds, we still may not have gotten the
		// entire string.  So when reads start succeeding, we have to
		// check the data we got for a null-terimated string.  If we didn't
		// get the full string, we change the byte count to either
		// move into the next higher range (if we were already reading
		// the max within the current range) or we set it to a factor
		// of 1.5 of the current byte count to try a value between the
		// current succeeding read and one that had failed.
		unsigned int loRead = 1;		// range boundaries
		unsigned int hiRead = 128;
		unsigned int cbRead = 128;		// number of bytes to read
		unsigned int chunkRead = 64;	// amount to change cbRead if we fail
										// we will not halve this before we read
		bool gotString = false;
		bool doneReading = false;
		while( !doneReading )
		{
			// try the read with the current byte count
			if( p->readDataSpace( pImageName, cbRead, buf, false ) )
			{
				// the read succeeded - 
				// did we get the full string?
				if( ev.u.LoadDll.fUnicode )
				{
					unsigned int cbReadIdx = cbRead / sizeof(WCHAR);
					wbuf[cbReadIdx] = L'\0';
					WCHAR* nulp = wcschr( wbuf, L'\0' );
					assert( nulp != NULL );			// because we just NULL-terminated the string
					gotString = (nulp != &(wbuf[cbReadIdx]));
				}
				else
				{
					buf[cbRead] = '\0';
					char* nulp = strchr( buf, '\0' );
					assert( nulp != NULL );			// because we just NULL-terminated the string
					gotString = (nulp != &(buf[cbRead]));
				}

				if( gotString )
				{
					doneReading = true;
				}
				else
				{
					// we didn't get the full string
					// we need to try a larger read
					if( cbRead == hiRead )
					{
						// we were at the high end of the current range -
						// move to the next range
						loRead = hiRead + 1;
						hiRead = loRead + 128 - 1;
						chunkRead = 128;				// we will halve this before we read again
						if( loRead > (MAX_PATH * sizeof(WCHAR)) )
						{
							// we've tried every range but still failed
							doneReading = true;
						}
						else
						{
							cbRead = hiRead;
						}
					}
					else
					{
						// we were within the current range -
						// try something higher but still within the range
						cbRead = cbRead + chunkRead;
					}
				}
			}
			else
			{
				// the read failed -
				// we need to try a smaller read
				if( cbRead > loRead )
				{
					unsigned int nextRead = cbRead - chunkRead;
					if( nextRead == cbRead )
					{
						// we can't subdivide any further
						doneReading = true;
					}
					else
					{
						cbRead = nextRead;
					}
				}
				else
				{
					// there are no smaller reads to try in this range,
					// and by induction, in any range.
					doneReading = true;
				}
			}

			// update the amount that we use to change the read request
			chunkRead /= 2;
		}

		if( !gotString )
		{
			// this is a serious problem because some read 
			// should've succeeded
			sprintf( msgText, "Failed to read DLL image name - no read succeeded\n" );
			logLine( msgText );
		}
		else
		{
			if( ev.u.LoadDll.fUnicode )
			{
				// the DLL path is a Unicode string
				// we have to convert it to single-byte characters
				char* tmpbuf = new char[MAX_PATH];

				WideCharToMultiByte(CP_ACP,		// code page to use (ANSI)
									0,			// flags
									wbuf,		// Unicode string
									-1,			// length of Unicode string (-1 => null-terminated)
									tmpbuf,		// destination buffer
									MAX_PATH,	// size of destionation buffer
									NULL,		// default for unmappable chars
									NULL);		// var to set when defaulting a char

				// swap buffers so that buf points to the single-byte string
				// when we're out of this code block
				delete[] buf;
				buf = tmpbuf;
			}
			ret = buf;
		}

		delete[] buf;
	}
	else
	{
		// we were given an image name pointer, but it was NULL
		// this happens for some system DLLs, and if we attach to
		// the process instead of creating it ourselves.
		// However, it is very important for us to know about kernel32.dll,
		// so we check for it specially.
		//
		// This call only changes the string parameter if the indicated file is
		// actually kernel32.dll.
		kludge_isKernel32Dll( ev.u.LoadDll.hFile, ret );
	}

	// cleanup
	delete[] msgText;

	return ret;
}

bool dyn_lwp::openFD()
{
  // Nothing to do here that I know of, since we are handed the FD
  // as part of a debug message
  return true;
}

void dyn_lwp::closeFD()
{
  return;
}
